% Ref: A. Del Bue, J. Xavier, L. Agapito, and M. Paladini, "Bilinear
% Factorization via Augmented Lagrange Multipliers (BALM)" ECCV 2010.
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; version 2, June 1991 
%
% USAGE: [S,M,t,timing,err_1] = balm_missing_bilinear(Y,D,Sinitial,Minitial,1,@proj_deformable_approx,100,10);
%
% This function solves for the general Augmented Lagrangian Multipliers
% problem. It uses a data matrix, a missing data mask and an initialisation of the
% bilinear components. The solution gives the new bilinear components
% satisfying the manifold constraints and the matrix completion for Y. The
% optimisation is standard for the given problem, the only thing varying is
% the projection given by the custom function "handle_projector".
%
% INPUT:
% Y: 2D measurement matrix storing each point trajectories at each column.
% the size of the matrix is 2F x P where F are the frames and P the points.
%
% S: the 3 x P shape matrix used for initialisation
%
% M: the 2F x 3 motion matrix used for initialisation
%
% req_cent: "1" compute the centroid, "0" assumes that the image points are
% centered to the image centroid.
%
% proj_handle: a function handle to the function that does the single frame
% projection.
% 
% Kmax: maximum number of outer loop iterations (check Alg.1 in the paper)
%
% Lmax: maximum number of inner loop iterations (check Alg.2 in the paper)
%
% OUTPUT:
%
% S: the 3 x P shape matrix containing the 3D coordinates of the shape
%
% M: the 2F x 3 motion matrix containing the camera projection matrices
%
% t: the 2F x 1 vector containing the 2D translation at each frame
% 
% timing: optimisation elapsed time
%
% err_1: error at each iteration defined as "err_1(k)= norm(Yprev -
% Yk,'fro')/numel(Yk);"

function [S,M,t,timing,err_1] = balm_factorization_bilinear(Y,S,M,req_cen,handle_projector,Kmax,Lmax)



[n,m] = size(Y);
r = size(M,1);

display('%%%%%%%%%%%%%%%%%%%%%%%%%%%  SFM  %%%%%%%%%%%%%%%%%%%%%%%%%%%')
p = 2;

C = null(ones(n,n));

% % with missing data non-rigid
% Kmax = 50;
% Lmax = 30;

% % with known data 
% Kmax = 50;

% % with missing data rigid
% Kmax = 10;
% Lmax = 6;

% display('%%%%%%%%%%%%%%%%%%%%%%%%%%%  PHOTOMETRIC STEREO  %%%%%%%%%%%%%%%%%%%%%%%%%%%')
% p = 1;
% 
% % with known data photometric
% Kmax = 25;
% Lmax = 15;

k = 0;
epsilon_best = +inf;
% sigma = 0.1;
% sigma = .1;
sigma = 50;
R = zeros(r,m);
gamma = 5;
eta = 0.5;




err_1 = zeros(1,Kmax);



if req_cen

    Sc = C'*S;
    Yc = C'*Y;

else
    Sc = S;
    Yc = Y;
    t = [];
end

% compute number of frames
f = m/p;

track_obj = [+inf];
track_epsilon = [+inf];



Yprev = Y;
Yk = Y;
% % instrumentation code
% track_obj = [];
% track_epsilon = [];
% track_yerror =[];

tic();
while k < Kmax
        
    Ytilde = sqrt(2/sigma)*Yc;
    
    l = 0;
    while l < Lmax
        
        %%%%%%% solve subproblem 1 %%%%%%%
        N = []; L = [];
        for i = 1:f
            Ni = handle_projector(M(:,(i-1)*p+1:i*p)-(1/sigma)*R(:,(i-1)*p+1:i*p));
            N = [ N , Ni ];

        end;                  
%         l
                                
        %%%%%%% solve subproblem 2 %%%%%%%
        
        S = Yc*pinv(M); 
        
        Stilde = sqrt(2/sigma)*S;                        
        C = N + (1/sigma)*R;
        Aux1 = [ Stilde ; eye(r) ];
        Aux2 = [ Ytilde ; C ];
        
        M = pinv(Aux1)*Aux2; 
           
       
        
        % update l
        l = l+1;
        
    end;

        

    eps_data = norm(Ytilde-Stilde*N,'fro')^2;
    epsilon = norm(M-N,'fro')^2;  
    track_epsilon = [ track_epsilon , epsilon];    
    track_obj = [ track_obj , eps_data ];   
 
    
% % % % instrumentation code
% 
%    figure(1);
%    clf;
%    plot(track_obj(2:end)); 
%    grid on;
% 
%    figure(2);
%    clf;
%    semilogy(track_epsilon(2:end));
%    grid on;
%     
% 
%   drawnow;    
    

  % Stop conditions
  if eps_data<1e-6  || epsilon<1e-7
      display('Converged');
      break
  end

%   if track_obj(end-1)< track_obj(end)
%       display('Local Minima');
%       N = Nold;
%       break
%   end  
  
  Nold = N;
   
        
    if epsilon < eta*epsilon_best
        R = R-sigma*(M-N);
        epsilon_best = epsilon;
    else
        sigma = gamma*sigma;
    end;

    
    % update k
    k = k+1;
    err_1(k)= norm(Yprev - Yk,'fro')/numel(Yk);
    Yprev = Yk;

    
    if rem(k,5) == 0
        fprintf('%2.2f%% done\n',k/Kmax*100);
    end;
    
end;
    
M = N;    
S = Y*pinv(M);

if req_cen
    t = (1/n)*ones(1,n)*(Y -S*M);
    t=t';
end

timing = toc();