% Ref: A. Del Bue, J. Xavier, L. Agapito, and M. Paladini, "Bilinear
% Factorization via Augmented Lagrange Multipliers (BALM)" ECCV 2010.
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; version 2, June 1991 
%
% USAGE: [S,M,t,timing,err_1] = balm_missing_bilinear(Y,D,Sinitial,Minitial,1,@proj_deformable_approx,100,10);
%
% This function solves for the general Augmented Lagrangian Multipliers
% problem. It uses a data matrix, a missing data mask and an initialisation of the
% bilinear components. The solution gives the new bilinear components
% satisfying the manifold constraints and the matrix completion for Y. The
% optimisation is standard for the given problem, the only thing varying is
% the projection given by the custom function "handle_projector".
%
% INPUT:
% Y: 2D measurement matrix storing each point trajectories at each column.
% the size of the matrix is 2F x P where F are the frames and P the points.
% Missing entries can be defined as a NaN
%
% Z: a 2F x P matrix with the missing entries. "0" is an unknown entry, "1"
% is a known entry.
%
% Sinitial: the 3 x P shape matrix used for initialisation
%
% Minitial: the 2F x 3 motion matrix used for initialisation
%
% req_cent: "1" compute the centroid, "0" assumes that the image points are
% centered to the image centroid.
%
% proj_handle: a function handle to the function that does the single frame
% projection.
% 
% Kmax: maximum number of outer loop iterations (check Alg.1 in the paper)
%
% Lmax: maximum number of inner loop iterations (check Alg.2 in the paper)
%
% OUTPUT:
%
% S: the 3 x P shape matrix containing the 3D coordinates of the shape
%
% M: the 2F x 3 motion matrix containing the camera projection matrices
%
% t: the 2F x 1 vector containing the 2D translation at each frame
% 
% timing: optimisation elapsed time
%
% err_1: error at each iteration defined as "err_1(k)= norm(Yprev -
% Yk,'fro')/numel(Yk);"

function [Sn,Mn,t,timing,err_1] = balm_missing_bilinear_alter(Y,Z,S,M,req_cent,handle_projector,Kmax,Lmax)

[n,m] = size(Y);
r = size(M,1);


p = handle_projector(); % dimensionality of the matrix to project
% p=1 for photometric stereo and p=2 for non-rigid SfM

k = 0;
epsilon_best = +inf;

sigma = 50;


R = zeros(r,m);
gamma = 5;
eta = 0.5;


err_1 = zeros(1,Kmax);




% compute number of frames
f = m/p;

track_obj = [+inf];
track_epsilon = [+inf];


% register to a shape centroid
ref_t=[];
Zhat=ones(n,m)-Z;
num_missing = sum(sum(Zhat));
if req_cent
    ref_t = find(sum(Z,1) == m, 1); % take away reference point
    ref_t = [];
    if ~isempty(ref_t)
        t = Y(ref_t(1),:)';
        Yc = Y - repmat(t',n,1);
    else
        [Yc,t]=balm_fact_register(Y);
    end
else
    t = [];
    Yc=Y;
end



S = Yc*pinv(M); 

Yprev = Y;
Yk = Y;
% % instrumentation code
% track_obj = [];
% track_epsilon = [];
% track_yerror =[];

tic();
while k < Kmax
        
    Ytilde = sqrt(2/sigma)*Yc;
    
    l = 0;
    while l < Lmax
        
        %%%%%%% solve subproblem 1 %%%%%%%
        N = []; L = [];
        for i = 1:f
            Ni = handle_projector(M(:,(i-1)*p+1:i*p)-(1/sigma)*R(:,(i-1)*p+1:i*p));
            N = [ N , Ni ];

        end;                  

                                
        %%%%%%% solve subproblem 2 %%%%%%%
        
        Stilde = sqrt(2/sigma)*S;                        
        C = N + (1/sigma)*R;
        Aux1 = [ Stilde ; eye(r) ];
        Aux2 = [ Ytilde ; C ];
        
        M = pinv(Aux1)*Aux2; 
        S = Yc*pinv(M);            
        
        if req_cent
            Yk=S*M.*Zhat + Yc.*Z + ones(n,1)*t'; 
            ref_t = []; % take away reference point
            if ~isempty(ref_t)
                t = Y(ref_t(1),:)';
                Yc = Y - repmat(t',n,1);
            else
                [Yc,t]=balm_fact_register(Yk);
            end    
        else
            t=[];
            Yc=S*M.*Zhat + Yc.*Z;
        end    
        
        Ytilde = sqrt(2/sigma)*Yc;    
        
        % update l
        l = l+1;
        
    end;

        

    eps_data = norm(Ytilde-Stilde*N,'fro')^2;

    epsilon = norm(M-N,'fro')^2;  
    track_epsilon = [ track_epsilon , epsilon];    
    track_obj = [ track_obj , eps_data ];   
 
    
% % % % instrumentation code
% 
%    figure(1);
%    clf;
%    plot(track_obj(2:end)); 
%    grid on;
% 
%    figure(2);
%    clf;
%    semilogy(track_epsilon(2:end));
%    grid on;
%     
% 
%   drawnow;    
    

  % Stop conditions
  if eps_data<1e-12  || epsilon<1e-13
      display('Converged');
      break
  end

%   if track_obj(end-1)< track_obj(end)
%       display('Local Minima');
%       N = Nold;
%       break
%   end  
  
  Nold = N;
     
    

    
    if epsilon < eta*epsilon_best
        R = R-sigma*(M-N);
        epsilon_best = epsilon;
    else
        sigma = gamma*sigma;
    end;

    
    % update k
    k = k+1;
    if req_cent
        err_1(k)= norm(Yprev - Yk,'fro')/numel(Yk);
        Yprev = Yk;
    else
        err_1(k)= norm(Yprev - Yc,'fro')/numel(Yc);
        Yprev = Yc;
    end
    
    

    
    if rem(k,5) == 0
        fprintf('%2.2f%% done\n',k/Kmax*100);
    end;
    
end;
    
Mn = N;    
Sn = Yc*pinv(M);
timing = toc();