% Ref: A. Del Bue, J. Xavier, L. Agapito, and M. Paladini, "Bilinear
% Factorization via Augmented Lagrange Multipliers (BALM)" ECCV 2010.
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; version 2, June 1991 
%
% USAGE: [Wfill,M,S,R,l,t,err_1,timing] = balm_missing_nrsfm(Y,D,5,@proj_deformable_approx,1000,10);
%
% This function solves for the Non-Rigid Structure from Motion problem. 
% Given a measurement matrix W and a missing data matrix, it computes the 2D 
% missing entries values (matrix completion), the 3D structure in the form 
% of a set of basis shapes, the 2D translation and the camera projection 
% matrix. Results may vary according to the selection of the outer and 
% inner loop parameters of the algorithm (Kmax and Lmax). If Kmax and Lmax 
% are not given, the algorithm uses the standard values as used in the 
% paper.
%
%
% INPUT:
% Y: 2D measurement matrix storing each point trajectories at each column.
% the size of the matrix is 2F x P where F are the frames and P the points.
% Missing entries can be defined as a NaN
%
% D: A 2F x P matrix with the missing entries. "0" is an unknown entry, "1"
% is a known entry.
%
% bases: a scalar defining the number of basis shapes
%
% proj_handle: a function handle to the function that does the single frame
% projection.
% 
% Kmax: maximum number of outer loop iterations (check Alg.1 in the paper)
% Lmax: maximum number of inner loop iterations (check Alg.2 in the paper)
%
% OUTPUT:
%
% Wfill: the new measurement matrix with the filled entries
%
% M: the 2F x 3 motion matrix containing the camera projection matrices
%
% S: the (3*bases x P) shape matrix containing the 3D basis shapes
%
% R: the 2F x 3 collection of orthographic camera matrices
%
% l: the (bases x F) matrix containing the configurarion weights for the
% basis shapes
%
% t: the 2F x 1 vector containing the 2D translation at each frame
% 
% err_1: error at each iteration defined as "err_1(k)= norm(Yprev -
% Yk,'fro')/numel(Yk);"
%
% timing: optimisation elapsed time

function [Wfill,M,S,R,l,t,err_1,timing] = balm_missing_nrsfm(Y,D,bases,proj_handle,Kmax,Lmax)


[F2,P]=size(Y);

%%%%%%%%%%%%%%%%%%% INITIALIZATION %%%%%%%%%%%%%%%%%%%
% Fill missing entries using rigid SfM with missing entries
% It is possible to reduce the amount of time used for initialisation by
% reducing the inner and outer iterations in balm_missing_rsfm.m

disp('start nonrigid initialisation')
if sum(sum(D))== F2*P % no missing data
    [Shape,Motion,T,scale] = fact_rigid(Y, 1);
else % missing data
    tic();
        Y(D==0)=nan;
        [Winit,Motion,Shape,Rstief,l,T,err_1,timing] = balm_missing_rsfm(Y,D,@proj_rigid_scaled_ortho,200,10);
        Y(D==0)=0;
    toc()
end

disp('done rigid completion')
[Winit,t]=register(Winit);

% initialise the basis shapes using the procedure suggested in "Nonrigid 
% structure-from-motion: Estimating shape and motion with hierarchical 
% priors", Torresani et al., PAMI 2008.

[L,B] = init_torresani(Winit, Motion, Shape, bases);
Minit = fact_plugLR(L',Motion);

disp('done nonrigid initialisation')

Winit = (Winit + t'*ones(1,P)).* not(D) + Y .* D;

timing=[];

if nargin < 5   
    Kmax=50;
    Lmax=20;
end

if nargin < 6   
    Lmax=1;
end

% do some cleaning before optimisation
clear Wn L B Rstief l err_1 timing scale T reproj
%%%%%%%%%%%%%%%%%%% BALM OPTIMISATION %%%%%%%%%%%%%%%%%%%

req_cent=1  % requires the computation of the umage centroid. Put 0 if already registered.

if sum(sum(D))== F2*P % no missing data
    [S,M,t,timing,err_1] = balm_factorization_bilinear(Winit',Shape',Minit',req_cent,proj_handle,Kmax,Lmax);
else % missing data
    [S,M,t,timing,err_1] = balm_missing_bilinear_alter(Winit',D',Shape',Minit',req_cent,proj_handle,Kmax,Lmax); 
end

%%%%%%%%%%%%%%% SEPARATE CAMERA MATRIX FROM DEFORMATIONS %%%%%%%%%%%%%%
S = S';
M = M';
[l, R] = fact_LRdecomposition(M); % separate camera from conf. weights
l=l';
Wfill = M*S + repmat(t,1,size(S,2)); % find the imputed matrix
Wfill = Wfill.*(ones(F2,P) - D) + Y.*D;