% Ref: A. Del Bue, J. Xavier, L. Agapito, and M. Paladini, "Bilinear
% Factorization via Augmented Lagrange Multipliers (BALM)" ECCV 2010.
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; version 2, June 1991 
%
% USAGE: [M,S,Ifill,err_1,timing] = balm_missing_photometric(I, D, Imask, @proj_photometric_4sph, Kmax, Lmax)
%
% This function solves for the Photometric Stereo problem with missing
% data. Given a set of images stored in I and a missing data matrix D, it
% computes the missing image pixels (matrix completion) and the 1st order
% spherical harmonics model as presented in: "R. Basri, D. Jacobs, I. 
% Kemelmacher, [1]"Photometric Stereo with General, Unknown Lighting" IJCV, 
% 2007". Initialisation for the missing pixels is made with the following 
% inpainting technique: "A. Criminisi, P. Perez, K. Toyama. Object Removal 
% by Exemplar-based Inpainting. CVPR, 2003". To run the code you need to
% download an implementation of this approach at:
% http://www.cc.gatech.edu/~sooraj/inpainting/ (check the code for more 
% info or email alessio.delbue@iit.it if you have troubles). Notice that
% the code only find the bilinear components M, S and the image completion. 
% To find the 3D shape given the normal you have to check [1, Sec. 4].
% Results may vary according to the selection of the outer and inner loop 
% parameters of the algorithm (Kmax and Lmax). If Kmax and Lmax are not 
% given, the algorithm uses the standard values as used in the paper.
%
%
% INPUT:
% I: Measurement matrix storing the image pixels at each frame the size of 
% the matrix is F x P where F is the number of frames and P the pixels.
% Each image is vectorized as a column i.e. P = img_h * img_w
%
% D: A F x P matrix with the missing entries. "0" is an unknown entry, "1"
% is a known entry.
%
% Imask: an image mask of size (img_h x  img_w) with a label "0" when a 
% pixel is in the background and "1" when it belongs to the shape.
%
% proj_handle: a function handle to the function that does the projection 
% at each pixel (normal constraints).
% 
% Kmax: maximum number of outer loop iterations (check Alg.1 in the paper)
%
% Lmax: maximum number of inner loop iterations (check Alg.2 in the paper)
%
% OUTPUT:
%
% Wfill: the new measurement matrix with the filled entries
%
% M: the 4 x P matrix storing the 1st order spherical harmonic (albedo + 
% normals). 
%
% S: the F x 4 matrix containing the lighting coefficients
%
% Ifill: the F x P matrix contains the optimised image pixels complying
% with an exact 1st order spherical harmonic model (in this respect the 
% reconstructed image will be dissimilar from the original on)
%
% err_1: error at each iteration defined as "err_1(k)= norm(Yprev -
% Yk,'fro')/numel(Yk);"
% 
% timing: optimisation elapsed time

function [M,S,Ifill,err_1,timing] = balm_missing_photometric(I, D, Imask, proj_handle, Kmax, Lmax)

%%%%%%%%%%%%%%%%%%% INITIALISATION %%%%%%%%%%%%%%%%%%%
% % % Initialisation with inpainting
Imask_col = Imask(:);
img_h = size(Imask,1);
img_w = size(Imask,2);
In = zeros(size(I));

disp('start photometric initialisation')
type1 = exist('inpaint.m');
type2 = exist('inpaint_gray.m');
if (type1 == 0) && (type2 == 0)
    display('First download the inpainting function')
    error('Inpainting function is missing please check http://www.cc.gatech.edu/~sooraj/inpainting/')
    % The inpaint function is made available by Sooraj Bhat and you need to
    % download the code (inpaint.m) from his webpage:
    % http://www.cc.gatech.edu/~sooraj/inpainting/
    % Notice that the function works for colour images so you have to
    % slightly modify his code if you are using grayscale images
end
    

for i = 1:size(I,1)
    i
    D_col = D(i,:)';
    Dreal = D_col + not(Imask_col);
    Dreal(Dreal>=1) = 1;
    Im_toproc =reshape(I(i,:)',img_h,img_w);
    Im_toproc(:,:,2)=Im_toproc;
    Im_toproc(:,:,3)=Im_toproc(:,:,1);
    D(i,:) = Dreal';
    % Call inpainiting

    
    [I_single,Morig,fillImg,Ctest,Dtest] = inpaint_gray...
        (Im_toproc,reshape(Dreal,img_h,img_w),[0 0 0]);
    I_single = rgb2gray(I_single);
    In(i,:) = I_single(:)';
end
clear Morig D_col Dreal I_single
disp('done photometric initialisation')

image_ind = find(Imask==1);
In           = In(:,image_ind);

[U, E, V]  = svds(In,4); % the 1st order harmonic has dimension 4
Lsvd        = U * sqrt(E(:,1:4));
Ssvd        = sqrt(E(1:4,:)) * V';


%%%%%%%%%%%%%%%%%%% NORMALIZATION %%%%%%%%%%%%%%%%%%%
mean_norm = mean([norm(Ssvd(2,:)), norm(Ssvd(3,:)), norm(Ssvd(4,:))]);
nscale     = diag([mean_norm/norm(Ssvd(1,:)), ...
                  mean_norm/norm(Ssvd(2,:)), ...
                  mean_norm/norm(Ssvd(3,:)), ...
                  mean_norm/norm(Ssvd(4,:))]);
              
Ssvd        = nscale * Ssvd;
Lsvd        = Lsvd * inv(nscale);

% % At the moment
% handle_projector = @proj_photometric_4sph;

%%%%%%%%%%%%%%%%%%% OPTIMISATION %%%%%%%%%%%%%%%%%%%

if nargin < 5   
    Kmax=100;
    Lmax=5;
end

if nargin < 6   
    Lmax=5;
end

req_cent=0; % no centroid computation in this case

[S,M,t,timing,err_1] = balm_missing_bilinear_alter(In,D(:,image_ind),Lsvd,Ssvd,req_cent,proj_handle,Kmax,Lmax);

Iless = S*M; % image completion
Ifill = zeros(size(I));
Ifill(:,image_ind) = Iless;
