% Ref: A. Del Bue, J. Xavier, L. Agapito, and M. Paladini, "Bilinear
% Factorization via Augmented Lagrange Multipliers (BALM)" ECCV 2010.
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; version 2, June 1991 
%
% USAGE: [Wfill,M,S,Rstief,l,t,err_1,timing] = balm_fact_missing_rsfm(Y,D,@proj_rigid_scaled_ortho,1000,10)
%
% This function solves for the Rigid Structure from Motion problem. Given a
% measurement matrix W and a missing data matrix, it computes the 2D 
% missing entries values (matrix completion), the 3D structure, the 2D
% translation and the camera projection matrix. Results may vary according
% to the selection of the outer and inner loop parameters of the algorithm
% (Kmax and Lmax). If Kmax and Lmax are not given, the algorithm uses the 
% standard values as used in the paper.
%
%
% INPUT:
% Y: 2D measurement matrix storing each point trajectories at each column.
% the size of the matrix is 2F x P where F are the frames and P the points.
% Missing entries can be defined as a NaN
%
% D: A 2F x P matrix with the missing entries. "0" is an unknown entry, "1"
% is a known entry.
%
% proj_handle: a function handle to the function that does the single frame
% projection.
% 
% Kmax: maximum number of outer loop iterations (check Alg.1 in the paper)
% Lmax: maximum number of inner loop iterations (check Alg.2 in the paper)
%
% OUTPUT:
%
% Wfill: the new measurement matrix with the filled entries. To notice that
% Wfill contains the original image trajectories in Y and the estimated
% missing data given by M,S and t 
% (i.e. Wfill = (M*S + repmat(t,1,size(S,2))).*not(D) + Y.*D)
%
% M: the 2F x 3 motion matrix containing the camera projection matrices
%
% S: the 3 x P shape matrix containing the 3D coordinates of the shape
%
% R: the 2F x 3 collection of orthographic camera matrices
%
% l: the 1 x F vector containing the scale for the orthographic cameras
%
% t: the 2F x 1 vector containing the 2D translation at each frame
% 
% err_1: error at each iteration defined as "err_1(k)= norm(Yprev -
% Yk,'fro')/numel(Yk);"
%
% timing: optimisation elapsed time

function [Wfill,M,S,R,l,t,err_1,timing] = balm_missing_rsfm(Y,D,proj_handle,Kmax,Lmax)

[F2,P]=size(Y);
Dhat=ones(F2,P)-D;

%%%%%%%%%%%%%%%%%%% INITIALIZATION %%%%%%%%%%%%%%%%%%%
% For each trajectory, fill the missing entries with the mean of the known 
% entries 
Y(D == 0) = 0;
Wk = (sum(Y,2)./sum(D,2)*ones(1,size(Y,2))).*Dhat + Y.*D; 

% % ref_t may be used if there is a trajectory (known) that can 
% % serve as 2D centroid. This option is disabled at the moment
% ref_t = find(sum(D,1) == F2, 1); % take away reference point
% ref_t = []; 
% num_missing = sum(sum(Dhat));
% if ~isempty(ref_t)
%     t = Y(:,ref_t(1))';
%     Wc = Wk - repmat(t',1,P);
% else
%     [Wc,t]=register(Wk);
% end

[Wc,t]=register(Wk);

[Sinit,Minit,T,scale_r] = fact_scaled_rigid(Wc); 
Sinit = scale_r*Sinit;
clear T scale

timing=[];

% do some cleaning before optimisation
clear  Wc
%%%%%%%%%%%%%%%%%%% BALM OPTIMISATION %%%%%%%%%%%%%%%%%%%

if nargin < 4   
    Kmax=1000;
    Lmax=10;
end

if nargin < 5   
    Lmax=1;
end

req_cent=1; % requires the computation of the umage centroid. Put 0 if already registered.

[S,M,t,timing,err_1] = balm_missing_bilinear_alter(Wk',D',Sinit',Minit',req_cent,proj_handle,Kmax,Lmax); 


S = S';
M = M';
[l, R] = fact_LRdecomposition(M); % separate camera from conf. weights
Wfill = M*S + repmat(t,1,size(S,2)); % find the imputed matrix
Wfill = Wfill.*Dhat + Y.*D;