%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; version 2, June 1991 
%
% USAGE: [L,B] = init_torresani(W2d, RR, S_rigid, rank)
%
% Compute the initialisation as described in Torresani et al. paper: 
% "Non-Rigid Structure-From-Motion: Estimating Shape and Motion with 
% Hierarchical Priors" PAMI 2008.
%
% INPUT:
%
% W2D: the 2D measurement matrix storing each point trajectories at each 
% column. The size of the matrix is 2F x P where F are the frames and P the
% points.
%
% RR: the (2F x 3) matrix containing the 2 x 3 orthographic camera matrices
% obtained from a rigid SfM initialisation.
% 
% S_rigid: the (3 x P) matrix containing the mean 3D shape obtained from
% rigid SfM initialisation.
%
% rank: the chosen number of basis shapes.
%
% OUTPUT:
%
% L: the F x r matrix containing the deformation weights.
% 
% B: the (3r x P) matrix containing the set of basis shapes.


function [L,B] = init_torresani(W2d, RR, S_rigid, rank)

[F, P] = size(W2d);
F = F/2;

% U-V model
U = zeros(3*F, rank);
V = zeros(rank, P);
% Transpose U-V model (Bregler's basis shapes)
L = zeros(F,rank);
B = zeros(3*rank,P);
K = rank;
W_tilda = register(W2d) - RR*S_rigid;
for k=1:K
   DATAfx3P = zeros(F, 3*P);
   for f=1:F,
      DATA3xP = pinv(RR([2*f-1 2*f],: ))*W_tilda(2*f-1:2*f,: );
      
      DATAfx3P(f,:) = DATA3xP(:)'; % Column-vectorization. [XYZ]_1,[XYZ]_2,...,[XYZ]_P
   end
   
   [basis,comp]=pca_decomposition(DATAfx3P,1); % rank-1 PCA: find direction of maximum variance.
   
   B(3*k-2:3*k,: ) = reshape(basis', 3, P); % basis shape
   L(:,k) = comp'; % shape coefficients

   
   for f = 1:F,       
      W_tilda(2*f-1:2*f,: ) = W_tilda(2*f-1:2*f,: ) - RR(2*f-1:2*f,: )*(L(f,k)*B(3*k-2:3*k,: )); % residual
   end   
end

B = [S_rigid;B];
L = [ones(F,1) L];