% Ref: A. Del Bue, J. Xavier, L. Agapito, and M. Paladini, "Bilinear
% Factorization via Augmented Lagrange Multipliers (BALM)" ECCV 2010.
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; version 2, June 1991 
%
% USAGE: Rart = proj_articulated(M)
%
% This function projects a generic matrix M of size 5 x 2 into the matrix
% Rart that satisfy the manifold constraints. This projection enforces the 
% articulated constraints given a hinge joint. This projector requires the
% CVX toolbox (http://cvxr.com/cvx/) installed. Details about the projector
% can be found in:  M. Paladini, A. Del Bue, S. M. s, M. Dodig, J. Xavier, and
% L. Agapito, "Factorization for Non-Rigid and Articulated Structure using
% Metric Projections" CVPR 2009.
% INPUT
%
% M: the 5 x 2 affine matrix 
%
% OUTPUT
%
% Rart: the 5 x 2 with manifold constraints

function Rart = proj_articulated(M)

if nargin < 1
    
    Rart = 2; % this says on how many columns of M the projection must be done
    return
end

[u,f]=relaxation_articulated(M(1,:)',M(2:3,:)',M(4:5,:)');
A = get_a(u,M(2:3,:)');
B = get_a(u,M(4:5,:)');
Rart = [u';A';B'];
end

function A=get_a(u,Y)
% returns A such that ||A-Y|| is minimized and u|A is stiefel
% find optimal A
P = sqrtm(eye(2)-u*u');
if ~ isreal(P) disp(u);error('p has imaginary');end
[U,D,V] = svd(P*Y);
Q = U*V';
A = P*Q;
end

function [u_relaxation,objective_score] = relaxation_articulated(x,Y,Z)
cvx_quiet(true);

% compute constants
R1 = Y*Y';
n1 = trace(R1);
d1 = abs(det(Y));

R2 = Z*Z';
n2 = trace(R2);
d2 = abs(det(Z));

I2 = eye(2);

aux = pinv([ I2(:) R1(:) R2(:) ])*kron(x,x); 
alfa = aux(1);
beta = aux(2);
gamma = aux(3);

% solve convex relaxation
cvx_begin
    
    variable U(2,2) symmetric;
        
    maximize(trace(U)+2*sqrt(alfa*trace(U)+beta*trace(R1*U)+gamma*trace(R2*U))+2*sqrt(n1-trace(R1*U)+2*d1*sqrt(1-trace(U)))+2*sqrt(n2-trace(R2*U)+2*d2*sqrt(1-trace(U))));
        
    U == semidefinite(2);
    trace(U) <= 1;
        
cvx_end;
[Uaux,Saux,Vaux] = svd(U);
u_relaxation = Uaux(:,1)*sqrt(Saux(1,1));
if x'*u_relaxation < 0
    u_relaxation = -u_relaxation;
end;



% extract solution
U = u_relaxation*u_relaxation';
objective_score = trace(U)+2*sqrt(alfa*trace(U)+beta*trace(R1*U)+gamma*trace(R2*U))+2*sqrt(n1-trace(R1*U)+2*d1*sqrt(1-trace(U)))+2*sqrt(n2-trace(R2*U)+2*d2*sqrt(1-trace(U)));
end